/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Input stream handling
 *
 * Provides usable elements from a stream, while abstracting subtleties like
 * salt string handling, missing ends-of-line, reading header, etc.
 *
 *
 * Stream varieties
 * ~~~~~~~~~~~~~~~~
 *
 *  Two kinds of streams can be read:
 *   - bare content of a test file
 *   - results from running the commands generated for this test file
 *
 *
 * Content of a result stream
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  The expected format for a result stream is a mix of salted strings and
 *  actual output lines. The former obeys the following format:
 *
 *    <salt> <rc> <keep> <skip> <prev_state><cur_state> @
 *
 *  where:
 *   - <salt> is a random string serving as differentiator with output lines
 *   - <rc> is the return code of the last command
 *   - <keep> is a number of lines to keep from the original test: commands
 *     and comments
 *   - <skip> is a number of lines of expected output that may be replaced
 *     with actual output
 *   - <prev_state> and <cur_state> are only informative, and give the latest
 *     parser state transition (* command, = output, # comment, . end)
 *   - '@' allows to quickly determine whether a line might be salted before
 *     actually parsing it
 *
 *  A salted line might be directly concatenated to the previous output line
 *  if this one didn't emit an end-of-line.
 *
 *
 *  The first line in a stream is a header:
 *
 *    <magic> <salt> <indent>
 *
 *  with:
 *   - <magic> a specific string confirming the nature of the stream
 *   - <salt> the salted string that will be used in subsequent lines
 *   - <indent> is the output indentation width
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "istream.h"



/* input stream */
struct istream {
	int type;             /* IS_TYPE_OUTPUT or IS_TYPE_TEST */

	FILE *fp;             /* input file pointer */

	unsigned int indent;  /* indentation */

	char *data;           /* read data */
	size_t datalen;       /* data length */

	unsigned int ahead;   /* flags of element stored ahead */
	size_t off;           /* offset of element from start of data */

	char *salt;           /* salt string if salted stream, or NULL */
	size_t saltlen;       /* salt string length */
};



/* --------------------------------------------------------------------------
 *   private functions
 */

static int
getnext_ahead(struct istream *is, struct line *l, unsigned int flags)
{
	char *start = is->data + is->off;

	is->off = 0;
	is->ahead = 0;

	if (flags & ISL_SALTED) {
		*start = *is->salt; /* restore first character */
		l->text = start + is->saltlen + 1;
	} else if (flags & ISL_OTHER) {
		l->text = start;
	} else {
		assert(0);
		return -1;
	}

	l->flags = flags;
	return 0;
}

static int
getnext_salt(struct istream *is, struct line *l)
{
	int ahead = is->ahead;
	char *psalt, *line;
	ssize_t n;

	if (ahead) /* data already available */
		return getnext_ahead(is, l, ahead);

	n = getline(&is->data, &is->datalen, is->fp);
	if (n == -1)
		return -1;

	line = is->data;
	if (n-- && line[n] == '\n') {
		line[n] = 0;
		if (n)
			n--;
	}

	psalt = (line[n] == '@') ? strstr(line, is->salt) : NULL;
	if (psalt && psalt != line) {
		is->off = psalt - line;
		is->ahead = ISL_SALTED;
		*psalt = 0; /* truncate */

		l->text = line;
		l->flags = ISL_OUTPUT | ISL_NO_EOL;
		return 0;
	}

	if (psalt) {
		l->text = psalt + is->saltlen + 1;
		l->flags = ISL_SALTED;
		return 0;
	} else {
		l->text = line;
		l->flags = ISL_OUTPUT;
		return 0;
	}
}

static void
extract_line_properties(char *line, ssize_t n, struct line *l)
{
	static const struct property {
		char last;
		const char * const string;
		unsigned int nb;
		unsigned int flag;
	} props[] = {
		{ 'e',  " (re)",      3,  ISL_REGEXP },
		{ 'l',  " (no-eol)",  7,  ISL_NO_EOL },
		{ 'b',  " (glob)",    5,  ISL_GLOB   },
	};
	int i;

	while (n > 0 && line[n--] == ')')
		for (i = 0; i < sizeof(props) / sizeof(*props); i++) {
			const struct property *p = &props[i];
			unsigned int nb = p->nb;

			if (line[n] == p->last                 &&
			    n > nb + 1                         &&
			    !strcmp(&line[n - nb], p->string))
			{
				n -= nb;
				line[n--] = 0;
				l->flags |= p->flag;
			}
		}
}

static int
getnext_nosalt(struct istream *is, struct line *l)
{
	ssize_t n;
	char *line, *p;

	n = getline(&is->data, &is->datalen, is->fp);
	if (n == -1)
		return -1;

	line = is->data;
	if (n-- && line[n] == '\n')
		line[n--] = 0;

	l->text = line;
	l->flags = 0; /* actual type is not checked, as the istream_getnext
	               * caller knows more about it than we do */

	p = line;
	if (p[n] == ')' &&                 /* might have properties */
	    *p++ == ' ' && *p++ == ' ' &&  /* might be an output */
	    *p   != '$' && *p   != '>')    /* is not a command */
		extract_line_properties(line, n, l);

	return 0;
}

static int
read_header_part(struct istream *is, int index, const char *str)
{
	int indent;

	switch (index) {
	case 0: /* magic sequence */
		if (!strcmp(str, "|klp|"))
			return 0;
		fprintf(stderr, "Invalid header\n");
		return -1;
	case 1:
		is->salt = strdup(str);
		if (!is->salt)
			return -1;
		is->saltlen = strlen(is->salt);
		return 0;
	case 2:
		indent = atoi(str);
		if (indent < 0) {
			fprintf(stderr, "Invalid indentation\n");
			return -1;
		}
		is->indent = (unsigned int)indent;
		return 0;
	default: /* open to future options/extensions */
		fprintf(stderr, "Unsupported header option: \"%s\"\n", str);
		return -1;
	}
}

static int
read_header(struct istream *is)
{
	char *line = NULL, *p, *str;
	size_t len = 0;
	ssize_t n;
	int i = 0, res = 0;

	n = getline(&line, &len, is->fp);
	if (n == -1) {
		if (!feof(is->fp))
			perror("getline");
		return -1;
	}

	p = line;
	while ( (res == 0) && (str = strsep(&p, " \n")) && *str)
		if (read_header_part(is, i++, str) == -1)
			res = -1;

	free(line);
	return res;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
istream_getnext(struct istream *is, struct line *l)
{
	return (is->salt) ? getnext_salt(is, l)
	                  : getnext_nosalt(is, l);
}

void
istream_push_back(struct istream *is, struct line *l)
{
	if (l->flags & ISL_SALTED) {
		is->off = l->text - is->data - is->saltlen - 1;
	} else if (l->flags & ISL_OTHER) {
		is->off = l->text - is->data;
	} else {
		assert(0);
		return;
	}

	is->ahead = l->flags;
}

unsigned int
istream_indent(struct istream *is)
{
	assert(is->salt);

	return is->indent;
}

void
istream_release(struct istream **is)
{
	if (!*is)
		return;

	free((*is)->data);
	free((*is)->salt);
	free(*is);
	*is = NULL;
}

int
istream_open_stream(struct istream *is)
{
	if (is->type == IS_TYPE_OUTPUT)
		return read_header(is);
	else
		return 0;
}

struct istream *
istream_alloc(int type, FILE *fp)
{
	struct istream *is;

	assert(type == IS_TYPE_TEST || type == IS_TYPE_OUTPUT);

	is = malloc(sizeof(*is));
	if (!is) {
		fprintf(stderr, "Out of memory.\n");
		return NULL;
	}

	is->type     = type;
	is->fp       = fp;
	is->indent   = 0;
	is->data     = NULL;
	is->datalen  = 0;
	is->ahead    = 0;
	is->off      = 0;
	is->salt     = NULL;
	is->saltlen  = 0;

	return is;
}
