/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Input stream handling
 */

#ifndef _ISTREAM_H
#define _ISTREAM_H

/* --------------------------------------------------------------------------
 *   input stream operations
 */

struct istream;
struct line;

enum {
	IS_TYPE_TEST,   /**< Test input stream */
	IS_TYPE_OUTPUT  /**< Shell output stream */
};

/**
 * Allocate memory for a new input stream
 *
 * \param[in] type  Type of stream
 * \param[in] fp    Associated file stream pointer
 *
 * \return          Handler suitable for istream operations,
 *                  or NULL in case of error
 */
struct istream *istream_alloc(int type, FILE *fp);

/**
 * Open the underlying stream
 *
 * For IS_TYPE_OUTPUT streams, a header will be read. Therefore, data
 * should be ready or the call will be blocking.
 *
 * \param[in] is  Handler returned by istream_alloc()
 *
 * \return        0 if the operation succeeded, -1 otherwise
 */
int istream_open_stream(struct istream *is);

/**
 * Release a stream
 *
 * \param[in,out] is  Pointer to istream handler (will be NULLed)
 */
void istream_release(struct istream **is);

/**
 * Query the indentation width of an output stream
 *
 * \param[in] is  Handler returned by istream_alloc()
 *
 * \return        Indentation width
 */
unsigned int istream_indent(struct istream *is);

/**
 * Get the next element/line in the stream
 *
 * \param[in]  is  Handler returned by istream_alloc()
 * \param[out] l   Line retrieved
 *
 * \return         0 if element successfully retrieved,
 *                -1 in case of error or end of stream
 */
int istream_getnext(struct istream *is, struct line *l);

/**
 * Push an element back into the stream
 *
 * \param[in]  is  Handler returned by istream_alloc()
 * \param[out] l   Line previously retrieved by istream_getnext()
 */
void istream_push_back(struct istream *is, struct line *l);



/* --------------------------------------------------------------------------
 *   istream_getnext() lines
 */

/* types */
#define ISL_SALTED  (1 << 0)  /**< Salted line       */
#define ISL_OUTPUT  (1 << 1)  /**< Shell output line */
#define ISL_OTHER   (1 << 2)  /**< Comment line      */

/* properties */
#define ISL_NO_EOL  (1 << 3)
#define ISL_REGEXP  (1 << 4)
#define ISL_GLOB    (1 << 5)

struct line {
	char *text;
	unsigned int flags;
};

#endif /* _ISTREAM_H */
