/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Merge a test output file with the originating test
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <assert.h>

#include "istream.h"
#include "merge.h"

struct output;
typedef void (*output_line_fn)(struct output *out,
                               char *line,
                               unsigned int flags);

struct output {
	FILE *fp;              /* file pointer for merged output */
	output_line_fn write;  /* write callback */
	bool verbatim;         /* disable escaping and regexp/glob support? */
	unsigned int indent;   /* indentation size */
	char *blank;           /* indentation buffer */
};

struct merge_context {
	struct istream *is_test;
	struct istream *is_result;
	struct output *out;
};


/* --------------------------------------------------------------------------
 *   private functions
 */

static void
output_amended_line(struct output *out, char *line, unsigned int flags)
{
	const char *p = line, *last = line;
	bool escaped = false;
	FILE *fp = out->fp;

	fwrite(out->blank, out->indent, 1, fp);

	if ((*p == '$' || *p == '>') && p[1] == ' ') {
		/* escape prefixes that would be interpreted
		 * as a command or command continuation */
		fputc('\\', fp);
		escaped = true;
	}

	while (1) {
		if (!isprint(*p)) {
			/* print the last series of printable bytes verbatim */
			if (p > last)
				fwrite(last, p - last, 1, fp);

			if (!*p)
				break;

			switch (*p) {
			case '\t': fputs("\\t", fp); break;
			case '\r': fputs("\\r", fp); break;
			case '\n': assert(0); break;
			default: fprintf(fp, "\\x%02x", *p & 0xFF); break;
			}

			last = p + 1;
		}

		p++;
	}

	if (last != line || escaped)
		fputs(" (esc)", fp);
	if (flags & ISL_REGEXP)
		fputs(" (re)", fp);
	if (flags & ISL_GLOB)
		fputs(" (glob)", fp);
	if (flags & ISL_NO_EOL)
		fputs(" (no-eol)", fp);

	fputc('\n', fp);
}

static void
output_verbatim_line(struct output *out, char *line, unsigned int flags)
{
	fwrite(out->blank, out->indent, 1, out->fp);
	fputs(line, out->fp);
	if (!(flags & ISL_NO_EOL))
		fputc('\n', out->fp);
}

static char *
glob_regexp(const char *src, char *dst, size_t len)
{
	const char *p = src;
	char *q = dst;

	while (q - dst < len - 2) {
		switch (*p) {
		case '\\':
			if (strchr("\\*?", p[1]))
				*q++ = *p;
			*q++ = *++p;
			break;
		case '*': *q++ = '.'; *q++ = '*'; break;
		case '?': *q++ = '.'; break;
		case 0: *q = 0; return dst;
		default: *q++ = *p; break;
		}
		p++;
	}

	fprintf(stderr, "Glob pattern too long.");
	return NULL;
}

static int
extended_match(const char *expect, const char *actual, unsigned int flags)
{
	const char *regexp = NULL;
	char buf[1024], err[256];
	regex_t preg;
	int res;

	if (flags & ISL_REGEXP)
		regexp = expect;
	else if (flags & ISL_GLOB)
		regexp = glob_regexp(expect, buf, sizeof(buf));

	if (!regexp)
		return -1;

	res = regcomp(&preg, regexp, REG_EXTENDED | REG_NOSUB);
	if (res) {
		regerror(res, &preg, err, sizeof(err));
		fprintf(stderr, "regcomp: %s: \"%s\"\n", err, regexp);
		return -1;
	}
	res = regexec(&preg, actual, 0, NULL, 0);

	return (res == REG_NOMATCH) ? -1 : 0;
}

static int
merge_inputs(struct istream *is_result,
             struct istream *is_test,
             unsigned int keep,
             unsigned int skip,
	     struct output *out)
{
	int is_rc;
	struct line l_test, l_result;
	char *expect, *actual;
	bool differs = false;
	unsigned int flags, res_flags;

	/* print lines from the test meant to be kept: commands and comments */
	while (keep--) {
		is_rc = istream_getnext(is_test, &l_test);
		if (is_rc == -1 || l_test.flags & ISL_SALTED) {
			fprintf(stderr, "Invalid input (keep)\n");
			return -1;
		}
		fprintf(out->fp, "%s\n", l_test.text);
	}

	/* attempt to match actual output with expected output */
	while (skip--) {
		if (istream_getnext(is_test, &l_test) == -1) {
			fprintf(stderr, "Invalid input (skip)\n");
			return -1;
		}

		if (differs || out->verbatim)
			continue; /* skip all remaining expected output */

		if (istream_getnext(is_result, &l_result) == -1) {
			differs = true;
			continue;
		}

		res_flags = l_result.flags;

		if (res_flags & (ISL_SALTED | ISL_OTHER)) {
			/* push string back into the stream */
			istream_push_back(is_result, &l_result);
			continue;
		}

		expect = l_test.text + 2;
		actual = l_result.text;

		if ((res_flags & ISL_NO_EOL) != (l_test.flags & ISL_NO_EOL) ||
		    (strcmp(expect, actual) /* no exact match */ &&
		     extended_match(expect, actual, l_test.flags) == -1))
		{
			differs = true;
			flags = l_result.flags & ~(ISL_REGEXP | ISL_GLOB);
			out->write(out, actual, flags);
			continue;
		}

		out->write(out, expect, l_test.flags);
	}

	/* print remaining output from result stream */
	while (istream_getnext(is_result, &l_result) != -1) {
		if (l_result.flags & (ISL_SALTED | ISL_OTHER)) {
			/* push string back into the stream */
			istream_push_back(is_result, &l_result);
			break;
		}

		flags = l_result.flags & ~(ISL_REGEXP | ISL_GLOB);
		out->write(out, l_result.text, flags);
	}

	return 0;
}

static int
cleanup(struct merge_context *c, int res)
{
	istream_release(&c->is_test);
	istream_release(&c->is_result);
	free(c->out->blank);
	free(c->out);
	free(c);

	return res;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
merge_init(struct merge_context *c)
{
	struct output *out = c->out;

	if (istream_open_stream(c->is_result) == -1 ||
	    istream_open_stream(c->is_test) == -1)
		return cleanup(c, -1);

	out->indent = istream_indent(c->is_result);

	out->blank = malloc(out->indent);
	if (!out->blank)
		return cleanup(c, -1);
	memset(out->blank, ' ', out->indent);

	return 0;
}

struct merge_context *
merge_alloc(FILE *fp_test,
            FILE *fp_result,
            FILE *fp_out,
            unsigned int flags)
{
	int verbatim = flags & MERGE_FLAG_VERBATIM;
	struct istream *is_test, *is_result;
	struct merge_context *c;
	struct output *out;

	is_test = istream_alloc(IS_TYPE_TEST, fp_test);
	if (!is_test)
		goto error_test;

	is_result = istream_alloc(IS_TYPE_OUTPUT, fp_result);
	if (!is_result)
		goto error_result;

	if ( !(c = malloc(sizeof(*c))) || !(out = malloc(sizeof(*out))) )
		goto error_oom;

	out->fp        = fp_out;
	out->verbatim  = verbatim;
	out->write     = verbatim ? output_verbatim_line
	                          : output_amended_line;
	out->blank     = NULL;

	c->out        = out;
	c->is_test    = is_test;
	c->is_result  = is_result;

	return c;

error_oom:
	fprintf(stderr, "Out of memory!\n");
	free(c);
error_result:
	istream_release(&is_test);
error_test:
	return NULL;
}

/* returns the number of result line processed (0 indicates end of input)
 * or -1 in case of error
 *
 * memory and input streams implicitly released when processing is over
 */
int
merge_process_result_line(struct merge_context *c)
{
	unsigned int keep, skip, rc;
	struct output *out = c->out;
	struct line l;
	int nn, res = -1;

	if (istream_getnext(c->is_result, &l) != -1) {
		if (!(l.flags & ISL_SALTED)) {
			fprintf(stderr, "Invalid input. Unexpected line.\n");
			goto out;
		}

		nn = sscanf(l.text, "%u %u %u", &rc, &keep, &skip);
		if (nn != 3) {
			fprintf(stderr, "Invalid input.\n");
			goto out;
		}

		if (rc)
			fprintf(out->fp, "  [%u]\n", rc);

		if (merge_inputs(c->is_result,
		                 c->is_test,
		                 keep,
		                 skip,
		                 out) == -1)
			goto out;

		return 1; /* processed one line */
	}
	res = 0; /* ok, end of input */

out:
	return cleanup(c, res);
}

int
merge_process_streams(FILE *fp_test,
                      FILE *fp_result,
                      FILE *fp_out,
                      unsigned int flags)
{
	struct merge_context *c;
	int res;

	c = merge_alloc(fp_test, fp_result, fp_out, flags);
	if (!c)
		return -1;

	if (merge_init(c) == -1)
		return -1;

	while ( (res = merge_process_result_line(c)) > 0)
		;

	return res;
}
