/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Shell spawning
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "fdutils.h"
#include "tmpdir.h"

extern pthread_mutex_t fdutils_mutex;


/* --------------------------------------------------------------------------
 *   public functions
 */

pid_t
start_shell(int *shfd_in, int *shfd_out, const char *name)
{
	int tdir_fd, pfd1[2], pfd2[2], fd_in, fd_out;
	pid_t pid;

	if ((tdir_fd = tmpdir_create(name)) == -1)
		return -1;

	if (pipe_cloexec(pfd1) == -1 || pipe_cloexec(pfd2) == -1)
		return -1;

	pthread_mutex_lock(&fdutils_mutex);
	pid = fork();
	pthread_mutex_unlock(&fdutils_mutex);
	switch (pid) {
	case -1:
		perror("fork");
		return -1;
	case 0: /* child */
		if (fchdir(tdir_fd) == -1) {
			perror("fchdir");
			_exit(1);
		}

		close(pfd1[1]);
		close(pfd2[0]);

		if (*shfd_in == -1)
			fd_in = pfd1[0];            /* allocate a new fd */
		else if (*shfd_in != STDIN_FILENO)
			fd_in = *shfd_in;           /* use the specified fd */
		else
			fd_in = -1;                 /* use stdin */

		if (fd_in != -1)
			if (dup2(fd_in, STDIN_FILENO) == -1) {
				perror("dup2 in");
				return -1;
			}
		close(pfd1[0]);

		if (*shfd_out == -1)
			fd_out = pfd2[1];           /* allocate a new fd */
		else if (*shfd_out != STDOUT_FILENO &&
		         *shfd_out != STDERR_FILENO)
			fd_out = *shfd_out;         /* use the specified fd */
		else
			fd_out = -1;                /* use stdout/stderr */

		if (fd_out != -1)
			if (dup2(fd_out, STDOUT_FILENO) == -1 ||
			    dup2(fd_out, STDERR_FILENO) == -1) {
				perror("dup2 out");
				return -1;
			}
		close(pfd2[1]);

		execl("/bin/sh", "sh", "-", NULL);

		fprintf(stderr, "exec: %s\n", strerror(errno));
		_exit(1);
	default: /* parent */
		close(pfd1[0]);
		close(pfd2[1]);
	}

	*shfd_in = pfd1[1];
	*shfd_out = pfd2[0];

	return pid;
}
