/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Handling of temporary test directories
 *
 * Upon start, a temporary directory is created, and all tests are executed
 * in a different directory thereof. Upon exit this temporary directory is
 * removed, unless otherwise specified via tmpdir_keep().
 *
 * It is also possible to specify a directory to use for tests through the
 * KUULTO_TMPDIR environment variable.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#define _XOPEN_SOURCE 700  /* nftw, mkdtemp */
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <ftw.h>
#include <errno.h>

#include "fdutils.h"
#include "tmpdir.h"

static pthread_mutex_t mkdir_mutex = PTHREAD_MUTEX_INITIALIZER;
static char tmpdir_buf[] = "/tmp/" PACKAGE "_XXXXXX";
static char *tmpdir;
static int tmpdir_fd = -1;
static int keep_on_exit;


/* --------------------------------------------------------------------------
 *   private functions
 */

static void
__attribute__ ((constructor))
init_tmpdir()
{
	const char *env = getenv("KUULTO_TMPDIR");

	if (env && *env) {
		tmpdir = strdup(env);
		if (!tmpdir) {
			fprintf(stderr, "Out of memory\n");
			exit(1);
		}

		/* not passed on to child processes */
		unsetenv("KUULTO_TMPDIR");
	} else if (!mkdtemp(tmpdir_buf)) {
		perror("mkdtemp");
		exit(1);
	} else {
		tmpdir = tmpdir_buf;
	}

	tmpdir_fd = open_cloexec(tmpdir, 0);
	if (tmpdir_fd == -1)
		exit(1);
}

static int
rm(const char *path, const struct stat *st, int info, struct FTW *pfwt)
{
	int (*rm_fn)(const char *path);

	switch (info) {
	case FTW_DP: rm_fn = rmdir; break;
	default: rm_fn = unlink; break;
	}

	if ((*rm_fn)(path) == -1)
		fprintf(stderr, "Could not remove %s: %s\n", path,
		                                             strerror(errno));

	return 0;
}

static void
__attribute__ ((destructor))
cleanup_tmpdir()
{
	close(tmpdir_fd);

	if (tmpdir != tmpdir_buf) {
		free(tmpdir);
		/* user-supplied directory - no removal */
	} else if (keep_on_exit) {
		fprintf(stderr, "Temporary directory kept: \"%s\"\n", tmpdir);
	} else {
		int flags = FTW_CHDIR | FTW_DEPTH | FTW_MOUNT | FTW_PHYS;
		if (nftw(tmpdir, rm, 16, flags) == -1)
			perror("Temporary directory cleanup");
	}
}

static const char *
get_test_directory_name(const char *name, char *buf, size_t len)
{
	struct stat st;
	int count = 0;

	if (fstatat(tmpdir_fd, name, &st, 0) == -1)
		return name;

	do {
		snprintf(buf, len, "%s.%d", name, ++count);
	} while (fstatat(tmpdir_fd, buf, &st, 0) != -1);

	return buf;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
tmpdir_create(const char *name)
{
	const char *dirname;
	char buf[PATH_MAX];
	int fd;

	pthread_mutex_lock(&mkdir_mutex);
	dirname = get_test_directory_name(name, buf, sizeof(buf));
	if ((mkdirat(tmpdir_fd, dirname, S_IRWXU)) == -1) {
		pthread_mutex_unlock(&mkdir_mutex);
		perror("mkdirat");
		return -1;
	}
	pthread_mutex_unlock(&mkdir_mutex);

	fd = openat_cloexec(tmpdir_fd, dirname, 0);
	if (fd == -1) {
		perror("testdir creation");
		return -1;
	}

	return fd;
}

void
tmpdir_keep()
{
	keep_on_exit = 1;
}
